import funcy, operator

import gaet.util          as util
import gaet.metrics.count as count

from typing     import Iterable
from gaet.types import Gene, MetricGroup, GenePattern


def is_membership_satisfied(genes: Iterable[Gene], products: Iterable[GenePattern]) -> bool:
    counts = count.all_product_counts(genes, products).values()
    return all(map(lambda x: x > 0, counts))

def membership_agreement(a: MetricGroup, b: MetricGroup) -> MetricGroup:
    return funcy.merge_with(lambda x: operator.eq(*x), a, b)

def singleton_metric_group(genes: Iterable[Gene]) -> MetricGroup:
    """
    Calculate group membership metrics for the gene group
    """
    f = lambda x: (str(x['group']), is_membership_satisfied(genes, x['products']))
    return dict(map(f, util.get_gene_patterns()))



import funcy
from typing     import Callable, Dict, Iterable, List
from gaet.types import Gene, MetricGroup

def calculate_group(
        asm: List[Gene],
        partition_f: Callable[[Iterable[Gene]], Dict[str, List[Gene]]],
        metric_f: Callable[[Iterable[Gene]], MetricGroup]
        ) -> Dict[str, MetricGroup]:
    """
    Given a partitioning function and metric calculation function, first partitions
    the assembly groups, then calculates the metrics for the genes in each group.
    """
    return funcy.walk_values(metric_f, partition_f(asm))

def calculate_group_difference(
        asm: List[Gene],
        ref: List[Gene],
        partition_f: Callable[[Iterable[Gene]], Dict[str, List[Gene]]],
        metric_f: Callable[[Iterable[Gene], Iterable[Gene]], MetricGroup]
        ) -> Dict[str, MetricGroup]:
    """
    Given a partitioning function and metric calculation function, first partitions
    the assembly and reference genes into groups, then calculates the metrics for
    the genes in each corresponding group.
    """
    asm_groups = partition_f(asm)
    ref_groups = partition_f(ref)
    return funcy.merge_with(lambda x: metric_f(x[0], x[1]), asm_groups, ref_groups)



import funcy

from typing     import Iterable, Callable, List, Dict, Tuple, Set
from functools  import reduce
from gaet.types import Gene, GenePattern

import gaet.util as util


def partition_genes_by_product(product_regexes: Set[GenePattern], genes: Iterable[Gene]) -> Dict[str, List[Gene]]:
    """
    Partitions a list of genes into groups based on regular expression matches to
    their product field.
    """
    def match_product(acc: List[Tuple[str, Gene]], g: Gene) -> List[Tuple[str, Gene]]:
        for p in list(product_regexes):
            if p.re.search(g.product):
                acc.append((p.name, g))
        return acc

    # Reduce list of genes into (group, gene) tuples then group into a dictionary
    # by the first field of the tuple.
    groups = dict(funcy.group_values(reduce(match_product, genes, [])))

    # Fill in any entries not found with an empty array
    for p in product_regexes:
        if p.name not in groups:
            groups[p.name] = []
    return groups


def partition_genes_by_type(genes: Iterable[Gene]) -> Dict[str, List[Gene]]:
    """
    Partitions a list of genes into groups based on their type. Returns a group
    named 'all' including all genes, excluding those not in the list of types in
    'gaet/gene_types/yml'.
    """
    whitelist = set(util.get_yaml_resource('gene_types')['gene_types'])
    whitelist.remove('all')
    is_gene_allowed = lambda gene: gene.type in whitelist
    whitelisted_genes = list(filter(is_gene_allowed, genes))
    groups = dict(funcy.group_by(lambda g: g.type, whitelisted_genes))
    groups['all'] = whitelisted_genes

    # Fill in any groups not found
    for group in whitelist:
        if group not in groups:
            groups[group] = []
    return groups



import sys, os.path, funcy
import ruamel.yaml as yaml

from typing     import Any, Dict, Optional, List
from functools  import partial
from gaet.types import GenePattern


def is_empty_gff_file(path: str) -> bool:
    with open(path, 'r') as f:
        return all(map(lambda x: x.startswith("#") or not x.strip(), f))

def does_contain_fasta(path: str) -> bool:
    with open(path, 'r') as f:
        return any(map(lambda x: '##FASTA' in x, f))



def check_gff_file(path: Optional[str]) -> str:
    """
    Ensures the given GFF file is valid
    """
    path = str(path)
    if not os.path.isfile(path):
        err_exit('missing_file', path = path)
    if is_empty_gff_file(path):
        err_exit('empty_gff_file', path = path)
    if not does_contain_fasta(path):
        err_exit('no_fasta_entry', path = path)
    return path


def get_yaml_resource(name: str) -> Any:
    """
    Loads the contents of a package YAML file
    """
    import pkgutil
    return yaml.load(pkgutil.get_data(__name__, name + '.yml').decode('UTF-8', 'ignore'))


def get_gene_patterns() -> List[Dict[str, List[GenePattern]]]:
    import re
    f = funcy.rpartial(
            funcy.update_in,
            ['products'],
            partial(map, lambda x: GenePattern(x['name'], re.compile(x['regex']))))

    return list(map(f, get_yaml_resource('gene_types')['membership_groups']))


def err_exit(err_name: str, **args: str) -> None:
    errors = get_yaml_resource('errors')
    sys.stderr.write(errors[err_name].format(**args))
    sys.exit(1)


def parse_args(args: List[str]) -> Dict[str, Optional[str]]:
    import argparse
    parser = argparse.ArgumentParser(description = "A gene-based genome assembly evaluation tool")
    parser.add_argument('assembly', help='A GFF file containing annotations for a genome assembly.')
    parser.add_argument('--reference', '-r', required = False, help='A GFF file containing annotations for a reference genome.')
    parser.add_argument('--output',    '-o', required = False, help='Location to write generated output metrics')
    return vars(parser.parse_args(args))


def destination(args: Dict[str, Optional[str]]) -> str:
    """
    Determines the location of the output file given the parsed input args
    """
    if 'output' in args:
        return str(args['output'])
    else:
        return 'gaet_metrics.yml'




from typing    import NamedTuple, Union, Dict, Optional
from typing.re import Pattern

Gene = NamedTuple('Gene',
        [('id', str), ('type', str), ('product', str), ('length', int), ('digest', int)])

Metric = Optional[Union[bool, int, float]]

MetricGroup = Dict[str, Metric]

# Named regular expression pulled from gaet/gene_types.yml
GenePattern = NamedTuple('GenePattern', [('name', str), ('re', Pattern)])



import funcy, operator

import gaet.util as util

from typing     import Iterable, Tuple, Dict, List
from functools  import partial
from gaet.types import MetricGroup, Gene, GenePattern

def count_gene_products(genes: Iterable[Gene], pattern: GenePattern) -> Tuple[str, int]:
    """
    For a given GenePattern, returns a count of all matching gene products in the
    list of genes.
    """
    product_names = list(map(lambda g: g.product, genes))
    f = funcy.rcompose(
            partial(map, pattern.re.search),
            partial(filter, funcy.identity),
            list,
            len,
            lambda x: (pattern.name, x))
    return f(product_names)


def all_product_counts(genes: Iterable[Gene], patterns: Iterable[GenePattern]) -> Dict[str, int]:
    """
    Given list of Genes, returns a Dict containing the counts of each GenePattern.
    """
    return dict(map(funcy.func_partial(count_gene_products, genes), patterns))


def gene_count_diff(a: MetricGroup, b: MetricGroup) -> MetricGroup:
    return funcy.merge_with(lambda x: operator.sub(*x), a, b)


def gene_count_metrics(genes: Iterable[Gene]) -> Dict[str, MetricGroup]:
    """
    Calculate the counts for each gene group
    """
    f = lambda x: (str(x['group']), all_product_counts(genes, x['products']))
    return dict(map(f, util.get_gene_patterns()))


def gene_group_difference_metrics(asm: List[Gene], ref: List[Gene]) -> Dict[str, MetricGroup]:
    """
    Calculate the difference in gene counts between the reference and the assembly
    """
    asm_counts = gene_count_metrics(asm)
    ref_counts = gene_count_metrics(ref)
    f = funcy.rcompose(
            partial(map, lambda x: gene_count_diff(*x)),
            partial(zip, asm_counts.keys()),
            dict)
    return f(zip(asm_counts.values(), ref_counts.values()))



import sys, os.path, funcy

import ruamel.yaml as yaml
import gaet.gff    as gff
import gaet.util   as util
import gaet.gene   as gene

import gaet.metrics.util       as metric_util
import gaet.metrics.size       as size
import gaet.metrics.count      as count
import gaet.metrics.membership as membership
import gaet.metrics.distance   as dist

from typing     import List, Dict, Optional, Tuple, Any
from gaet.types import MetricGroup, Gene
from funcy      import partial, rpartial


def run() -> str:
    args = util.parse_args(sys.argv[1:])

    assembly     = list(gff.parse_gff_into_genes(util.check_gff_file(args['assembly'])))
    gaet_metrics = {'assembly' : simple_metrics(assembly)}

    if ('reference' in args) and args['reference']:
        reference = list(gff.parse_gff_into_genes(util.check_gff_file(args['reference'])))
        gaet_metrics['reference']  = simple_metrics(reference)
        gaet_metrics['comparison'] = comparison_metrics(assembly, reference)

    dst  = util.destination(args)
    with open(dst, "w") as f:
        doc = "---\n" + yaml.dump(gaet_metrics, default_flow_style = False)
        f.write(doc)
    return ""


def simple_metrics(genes: List[Gene]) -> Any:
    patterns   = set(funcy.mapcat(lambda x: x['products'], util.get_gene_patterns()))
    by_product = partial(gene.partition_genes_by_product, patterns)

    f = partial(metric_util.calculate_group, genes)

    return {"gene_type_size"   : f(gene.partition_genes_by_type, size.group_size),
            "gene_copy_size"   : f(by_product, size.group_size),
            "minimum_gene_set" : membership.singleton_metric_group(genes)}


def comparison_metrics(asm: List[Gene], ref: List[Gene]) -> Any:
    patterns   = set(funcy.mapcat(lambda x: x['products'], util.get_gene_patterns()))
    by_product = partial(gene.partition_genes_by_product, patterns)

    f = partial(metric_util.calculate_group_difference, asm, ref)

    return {'gene_type_distance' : f(gene.partition_genes_by_type, dist.distance_metrics),
            'gene_type_size'     : f(gene.partition_genes_by_type, size.group_size_difference),
            'gene_copy_size'     : f(by_product, size.group_size_difference),
            'gene_copy_distance' : f(by_product, dist.distance_metrics),
            'gene_set_agreement' : membership.membership_agreement(
                                        membership.singleton_metric_group(asm),
                                        membership.singleton_metric_group(ref))}



import funcy

from collections import defaultdict
from math        import sqrt
from typing      import Any, Iterable, Optional, Set, List, Union
from gaet.types  import Gene, MetricGroup

import gaet.util as util
import gaet.gene as gene


def gene_digest_set(genes: Iterable[Gene]) -> Set[int]:
    return set(map(lambda x: x.digest, genes))

def intersect(a: Iterable[Gene], b: Iterable[Gene]) -> int:
    return len(gene_digest_set(a) & gene_digest_set(b))

def difference(a: Iterable[Gene], b: Iterable[Gene]) -> int:
    return len(gene_digest_set(a) - gene_digest_set(b))

def symmetric_difference(a: Iterable[Gene], b: Iterable[Gene]) -> int:
    return len(gene_digest_set(a) ^ gene_digest_set(b))

def divide(x: int, y: int) -> Optional[float]:
    if y == 0:
        return None
    else:
        return x / float(y)

def norm(a: Iterable[Gene], b: Iterable[Gene], p) -> Union[int, float]:
    counts = defaultdict(lambda: 0)

    for gene in a:
        counts[gene.digest] += 1

    for gene in b:
        counts[gene.digest] -= 1

    if p == 1:
        return sum(map(abs, counts.values()))
    else:
        sqr = lambda x: x**2
        return round(sqrt(sum(map(sqr, counts.values()))), 3)


def distance_metrics(assembly: Iterable[Gene], reference: Iterable[Gene]) -> MetricGroup:
    n_intersect = intersect(assembly, reference)
    n_sym_diff  = symmetric_difference(assembly, reference)
    n_asm       = difference(assembly, reference)
    n_ref       = difference(reference, assembly)
    total       = n_intersect + n_sym_diff

    return {'n_intersect'               : n_intersect,
            'n_assembly_only'           : n_asm,
            'n_reference_only'          : n_ref,
            'n_symmetric_difference'    : n_sym_diff,
            'perc_intersect'            : divide(n_intersect, total),
            'perc_assembly_only'        : divide(n_asm, total),
            'perc_reference_only'       : divide(n_ref, total),
            'perc_symmetric_difference' : divide(n_sym_diff, total),
            'l1_norm'                   : norm(assembly, reference, 1),
            'l2_norm'                   : norm(assembly, reference, 2)}



from typing import Iterable, Optional, TypeVar

T = TypeVar('T')

def ffirst(xs: Iterable[Iterable[T]]) -> Optional[T]:
    realised = list(xs)
    if realised:
        return list(realised[0])[0]
    else:
        return None

def identity(x: T) -> T:
    return x



import funcy
from typing          import Iterable, Dict, TypeVar, Callable, Optional, List, Union
from functools       import partial
from gaet.types      import Metric, MetricGroup, Gene
from gaet.functional import identity, ffirst

import gaet.util as util
import gaet.gene as gene

SizeTransform = Callable[[Iterable[int]], Iterable[int]]


def calculate(f: SizeTransform, threshold: Union[int, float], xs: Iterable[int]) -> Optional[int]:
    """
    Calculate size-related metrics. Given a list of integers, select the first value
    where the first value from the running cumulative sum of the ordered list of
    values is greater than the given threshold value.

    Keyword Arguments:

    f         -- a function taking a list of integers and returning a list of
                 integers. Used to transform the integer list when calculating the
                 threshold metric.

    threshold -- integer at which to return the first transformed value for which
                 the running cumulative is greater than this.

    xs        -- list of integers to calculate this metric for
    """
    sorted_xs     = sorted(xs, reverse = True)
    trsfm   = zip(f(sorted_xs), funcy.sums(sorted_xs))
    value = ffirst(funcy.dropwhile(lambda x: x[1] < threshold, trsfm))
    if value:
        return int(value)
    else:
        return None


def nx(xs: Iterable[int], threshold: Union[int, float]) -> Optional[int]:
    """
    Calculate the size metric using the identity transform, e.g. ng50.
    """
    return calculate(lambda x: map(identity, x), threshold, xs)


def lx(xs: Iterable[int], threshold: Union[int, float]) -> Optional[int]:
    """
    Calculate the size metric using the value's index. e.g. lg50.
    """
    increment_index = lambda y: y[0] + 1
    f = lambda x: map(increment_index, enumerate(x))
    return calculate(f, threshold, xs)


def gene_lengths(genes: Iterable[Gene]) -> int:
    return list(map(lambda x: x.length, genes))


def metric_difference(a: MetricGroup, b: MetricGroup) -> MetricGroup:
    """
    Calculate the difference in two metric groups by subtracting the reference
    metrics from the assembly metrics.
    """
    def _f(metrics: List[Metric]) -> Metric:
        if any(map(lambda x: x is None, metrics)):
            return None
        else:
            return metrics[0] - metrics[1]
    return funcy.walk_values(_f, funcy.merge_with(list, a, b))


def group_size(genes: Iterable[Gene]) -> MetricGroup:
    """
    Calculate the set of size metrics for a given list of genes
    """
    sizes = gene_lengths(genes)
    sum_length = sum(sizes)
    half_assembly = sum_length / 2.0
    return {"sum_length" : sum_length,
            "count"      : len(sizes),
            "n50"        : nx(sizes, half_assembly),
            "l50"        : lx(sizes, half_assembly)}


def group_size_difference(asm: Iterable[Gene], ref: Iterable[Gene]) -> MetricGroup:
    """
    Calculate the set of size metrics for a given list of genes, using the length of
    reference genes as the threshold value.
    """
    sizes = gene_lengths(asm)
    half_ref_assembly = sum(gene_lengths(ref)) / 2.0
    reference_metrics = {"ng50" : nx(sizes, half_ref_assembly),
                         "lg50" : lx(sizes, half_ref_assembly)}
    difference_metrics = metric_difference(group_size(asm), group_size(ref))
    return funcy.merge(reference_metrics, difference_metrics)



import pkg_resources, os, pkgutil
from setuptools import setup, find_packages
from gaet.version import __version__

def read(fname):
    file_ = pkg_resources.resource_filename(__name__, fname)
    with open(file_, 'r') as f:
        return f.read()

setup(
    name                 = 'gaet',
    version              = __version__,
    description          = "Evaluate the quality of a genome assembly based on annotated genes",
    long_description     = read('README.rst'),
    author               = 'Michael Barton',
    author_email         = 'mail@michaelbarton.me.uk',
    install_requires     = read('requirements/default.txt').splitlines(),
    license              = "BSD",
    keywords             = "genomics bioinformatics sequencing assembly",
    url                  = "https://gitlab.com/michaelbarton/gaet/",
    scripts              = ['bin/gaet'],
    packages             = find_packages(),
    package_data         = {'': ['gene_types.yml']},
    include_package_data = True,
    classifiers          = [
        "License :: OSI Approved :: BSD License",
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.5',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
)



import gffutils, pyfaidx, hashlib, tempfile

from typing     import Iterable
from gaet.types import Gene

import gaet.util as util

def gff_to_fasta(path: str) -> pyfaidx.Fasta:
    _, tmp_file = tempfile.mkstemp(suffix = '.fasta', prefix = 'gaet_')
    with open(tmp_file, 'w') as fasta:
        with open(path, 'r') as gff:
            seq_flag = False
            for line in gff:
                if seq_flag:
                    fasta.write(line)
                if line.strip() == '##FASTA':
                    seq_flag = True
    return pyfaidx.Fasta(tmp_file)


def create_digest(s):
    return int(hashlib.sha256(s.encode('utf-8')).hexdigest(), 16)


def seq_digest(fasta: pyfaidx.Fasta, f: gffutils.Feature) -> int:
    seq = f.sequence(fasta, use_strand = False)
    if f.strand == "-":
        seq = seq.reverse.complement
    return create_digest(seq.seq)


def determine_product(f: gffutils.Feature) -> str:
    attr_keys = f.attributes.keys()
    if f.featuretype == 'repeat_region':
        return f['rpt_family'][0]
    elif 'product' in attr_keys:
        return f['product'][0]
    else:
        return 'hypothetical protein'


def feature_to_gene(fasta: pyfaidx.Fasta, f: gffutils.Feature) -> Gene:
    return Gene(f.id, f.featuretype, determine_product(f), f.stop - f.start + 1, seq_digest(fasta, f))


def parse_gff_into_genes(path: str) -> Iterable[Gene]:
    try:
        gff = gffutils.create_db(path, dbfn=':memory:').all_features()
    except ValueError:
        util.err_exit('gff_parsing_err', path = path)
    fasta = gff_to_fasta(path)
    return map(lambda x: feature_to_gene(fasta, x), gff)
